/*
 *      BKHTTP.C
 *
 *  Background routines for HTTP
 *
 *  This is the state machine which
 *  handles all Http transactions
 *
 *--*
 *      Originally written for:
 *        TCP/IP kernel for NCSA Telnet
 *          by Bob Fillmore & John Lydic
 *
 *      National Center for Supercomputing Applications
 *      152 Computing Applications Building
 *      605 E. Springfield Ave.
 *      Champaign, IL  61820
 *
 *      Converted to WatTcp kernel
 *
 *--*
 *  Revision history:
 *
 *  6/99   Separated from main into individual bkgr_xxx.c files rrf
 *  8/99   Cleanup for conversion to Waterloo TCP (WATTCP)  jwl/rrf
 *         Conversion for encapsulated state machine        jwl
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <alloc.h>
#include <bios.h>
#include <dos.h>
#include "whatami.h"
#include "http.h"
#include "bghttp.h"
#include <wattcp.h>
#include "LedCtrl.h"

/*
*  Defines
*/

#ifdef MSC
#define O_RAW  O_BINARY
#ifdef __TURBOC__
#include <dir.h>
#else
#include <direct.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <io.h>
#endif  /* msc */

// Function prototypes

void  setLedState(int State);
int ProcessPost(char *RefStr, char *DataButtonStr);
char *ProcessOutputString(char *theHttpString);
void Send404Response(char *Name404);
void NetWrite(char *theBuffer);
NoDots(char *filePath);

/*
*  Global Variables
*/

static pathname[PATHLEN],  		/* space to keep path names */
  newfile[PATHLEN],   			/* current file being received */
  myuser[17],         			/* user name on my machine */
  mypass[30],         			/* user pass on my machine (blank unless anonftp) */
  hisuser[17],        			/* user name on his machine */
  waitchar;           			/* character waiting for from net */

static int
  curstate = -1,   			/* state machine for background process */
  retstate = 200,  			/* to emulate a subroutine call */
  httpstate = 0,   			/* state of the http data transfer */
  isdir = 0,       			/* flag for rcp target pathname */
  waitpos = 0,     			/* marker for gathering strings from net */
  cnt = 0,         			/* number of characters from last netread()*/
  fh = 0,          			/* file handle when transfer file is open */
  httpfh = 0,      			/* file handle for ftp data */
  xp = 0,          			/* general pointer */
  towrite = 0,     			/* file transfer pointer */
  len = 0;         			/* file transfer length */

/* http methods */
char *methods[METHODS] = { "GET",
                           "HEAD",
                           "POST",
                           "PUT",
                           "DELETE",
                           "SECURE",
                           "LINK",
                           "UNLINK" };

/* header messages */
#define HEADERS 11
char *headers[HEADERS] = { "no match:",
                           "referer:",
                           "connection:",
                           "user-agent:",
                           "host:",
                           "accept:",
                           "accept-encoding:",
                           "accept-language:",
                           "accept-charset:",
                           "content-type:",
                           "content-length:" };

/* File types */

#define HTML_FILE       1
#define JPEG_FILE       2
#define GIF_FILE        3
#define OTHER_FILE      100

char *FileSuffix[6] = {     "htm",
                            "html",
                            "jpeg",
                            "jpe",
                            "jpg",
                            "gif" };

int FileType[6] =       {   HTML_FILE,
                            HTML_FILE,
                            JPEG_FILE,
                            JPEG_FILE,
                            JPEG_FILE,
                            GIF_FILE };
#define MaxTypes    6

//
//  Embedded SSDI function stuff
//
char *monthTbl[13] = { "NoMonthZero", "January", "February", "March",
                       "April", "May", "June",
                       "July", "August", "September",
                       "October", "November", "December" };

void Scr1Dec3();    // 3 Decimal digits data from Com1: Scratch Register
void Scr1Hex2();    // 2 Hex digits  data from Com1: Scratch Register
void Scr1Bin8();    // 8 binary digits data from Com1: Scratch Register
void Time();        // hh:mm:ss.ss
void Date1();       // mm/dd/yyyy
void Date2();       // mmmmmmmm dd,yyyy  i.e. January 1, 1999
void DateRfc1123();    // RFC 1123 compliant date function
void Scr1Hex2I();   // 2 Hex digits inverted data from Com1: Scratch Register
void Scr1Bin8I();   // 8 binary digits inverted data from Com1: Scratch Register

struct ssdiEntry { char *tag;        // The tag
                   char *fmt;        // The fmt
                   void (*func)();   // This operation function
};

struct ssdiEntry ssdiTable[] = {
	{ "scratch1", "decimal", Scr1Dec3 },  // <!--#exec scratch1=decimal -->
	{ "scratch1", "hex", Scr1Hex2 },      // <!--#exec scratch1=hex -->
	{ "scratch1", "hexi", Scr1Hex2I },    // <!--#exec scratch1=hexi -->
	{ "scratch1", "binary", Scr1Bin8 },   // <!--#exec scratch1=binary -->
	{ "scratch1", "binaryi", Scr1Bin8I }, // <!--#exec scratch1=binaryi -->
	{ "time", "time", Time },             // <!--#exec time=time -->
	{ "date1", "date1", Date1 },          // <!--#exec date1=date1 -->
	{ "date2", "date2", Date2 },          // <!--#exec date2=date2 -->
	{ 0, 0, 0 }
		};

static long int filelen = 0L;  /* length of current file for transfer */

static char lpname[30];

#define MaxReadLength 1024
#define DISCARD 1
#define NODISCARD 0

static char HoldBuf[8192];  // Buffer into which entire output page is built
static char *HoldPtr;

static char MungBuf[MaxReadLength];  // Buffer to read temp connection data
static char *ParsePtr;

static char crfound = 0;
static char HoldBuf[8192];

static char TempBuf[2046];
static char *HeaderLine[100];
static int holdtot, lines, BodySize, residual, NextHeadCode = HEAD_MAX+1;
static int ReqFileType;
static char body;
static char *nextfile;    /* pointer to the next filename returned by DOS */

tcp_Socket HttpSocket;   /* socket control structure */

static int Sfread(int, char *, int);

/* function prototypes*/

char *GetHeaderLine(char *BufferPtr, char *LineBuf, int *count, int MaxLength);
int ParseHeader( char *buffer);
int SendResponseFile(char *Name404, char *filename, int HeaderOnlyFlag);
int SendNormalFile(char *Name404, char *filename, int HeaderOnlyFlag);
int GetMethodCode(char *buffer);
void GenerateFileName(char *FileName, char *path);
void ParsePostRequest(char *PostBuffer, int BytesRead);
int GetFileType(char *buffer);

/* HTTPD stuff starts here */

/**
 **  InitHttpd - initialize Http state machine
 **/

void InitHttpd(void )
{
	char time[64], date[64];

	httpstate = IdleStartListen;          /* waiting for connection */
	Date2(date);
	Time(time);

	printf(VERSIONINFO);
	printf("\nStarting HTTP daemon on %s at %s\n", date, time);
}

/**
 **  Name: httpd
 **
 **  Function: Accept an incoming http request and service it.
 **
 **  Binding:  void httpd()
 **
 **  Returns:  void
 **/

void httpd()
{
	char FileName[200], LineBuf[200], TmpBuf[25];
	struct sockaddr CurConnectionAddress;
	
	int len, CharCount;
	int mcode, hdrOnly;
	char *method, *path, *ptr;
	REQUEST *aRequest;
	long remoteHostAddress = 0;      /* any host can connect */
	unsigned short remotePort = 0;   /* any remote port */
	unsigned short timeout = 0;      /* no timeout */

	switch (httpstate)
	{
		default:
			break;

		case IdleStartListen:      // setup to listen for an HTTP connection
			tcp_listen((void *) &HttpSocket,
				HttpPort, remoteHostAddress, remotePort, NULL, timeout);

		case IdleWaitConnect:      // wait for a connection to the port
			httpstate = IdleWaitConnect;

			// wait for connection
			while (sock_established((void *) &HttpSocket) == 0)
				break;   // Allow other work while waiting for connection

			// Set No Nagle mode
			// Data is sent as it is written with no PAD buffering
			sock_mode((void *) &HttpSocket, TCP_MODE_NONAGLE);

		case IdleRead:           // Connection established, perform read
			httpstate = IdleRead;

			// Read data
			if ((len = sock_fastread((void *) &HttpSocket,
				MungBuf, MaxReadLength)) == -1)
			{
				httpstate = IdleStartListen;   // If connection closed - re-open it
				break;
			}

			residual = 0; /* no leftovers in idle state */

			if (len > 0)
			{
				// connection started
				GetHeaderLine(MungBuf, LineBuf, &CharCount, len);
				
				// logging info in Common Logfile Format
				getpeername((void *) &HttpSocket, &CurConnectionAddress, NULL);
				printf("%s ",inet_ntoa(TmpBuf, CurConnectionAddress.s_ip));
				DateRfc1123(&TmpBuf);
				printf(" - - [%s] \"%s\"\n", TmpBuf, LineBuf);

				method = ptr = &LineBuf[0];
				while (*ptr != 0x20)     /* find first space */
					ptr++;
				*ptr++ = 0;                        /* terminate method string */
				mcode = GetMethodCode(method);

				switch (mcode)
				{
					case M_HEAD:
					case M_GET:
						path = ptr;            /* path string is next */
						while (*ptr != 0x20)   /* next space */
							ptr++;
						*ptr++ = 0;            /* terminate path string */
						GenerateFileName(FileName, path);

  						hdrOnly = mcode == M_HEAD ? 1 : 0;

						if (ReqFileType == HTML_FILE)
							SendResponseFile(path, FileName, hdrOnly);
						else
							SendNormalFile(path, FileName, hdrOnly);

						httpstate = WaitClose0;
						break;

					case M_POST:
					case M_PUT:
						httpstate = RecordPut;
						HoldPtr = &HoldBuf[0];
						ParsePtr =  &HoldBuf[0];
						ParsePostRequest(MungBuf, len);
						break;

					default:
						break;
				}
			}
			break;

		case WaitClose0:   // Flush the socket, wait until empty
			sock_flush((void *) &HttpSocket);  // Flush the output

			if (!sock_tbused((void *) &HttpSocket))  // IF: socket empty
			{
				sock_close((void *) &HttpSocket);     // Close the connection
				httpstate = WaitClose1;     // Wait for closed connection
			}
			break;

		case WaitClose1:    // Wait until connection actually closed
			if (tcp_tick((void *) &HttpSocket) == 0)
				httpstate = IdleStartListen;   // Go Idle Await Next Connection
			break;

		case RpStartListen:   // Connection dropped during put
			// setup the connection
			tcp_listen((void *) &HttpSocket,
				HttpPort, remoteHostAddress, remotePort, NULL, timeout);

		case RpWaitConnect:   // Await re-connection
			httpstate = RpWaitConnect;
			// wait for connection
			while (sock_established((void *) &HttpSocket) == 0)
				break;     // Allow other work while awaiting connect

			// Set No Nagle mode
			// Data is sent as it is written with no PAD buffering
			sock_mode((void *) &HttpSocket, TCP_MODE_NONAGLE);
			break;

		case RecordPut:   // read a PUT record
			if ((len = sock_fastread((void *) &HttpSocket,
				MungBuf, MaxReadLength)) == -1)  // Read data
			{
				httpstate = RpStartListen;   // If connection closed - re-open it
				break;
			}

			if (len > 0)
				ParsePostRequest(MungBuf, len);
			break;
	}
}

/**
 **  Name: GetMethodCode
 **
 **  Function: return a Method Code (Index) based
 **            on the specific HTTP method
 **
 **  Binding:  int GetMethodCode(char *buffer)
 **
 **            buffer - text buffer containing method
 **
 **  Returns:  Index for switch statement based
 **            on ASCII method code
 **/

int GetMethodCode(char *buffer)
{
	int mcode;

	if (!strcmp(buffer, methods[M_HEAD]))
		mcode = M_HEAD;
	else if(!strcmp(buffer, methods[M_GET]))
		mcode = M_GET;
	else if(!strcmp(buffer, methods[M_POST]))
		mcode = M_POST;
	else if(!strcmp(buffer, methods[M_PUT]))
		mcode = M_PUT;
	else
		mcode = M_INVALID;

	return mcode;
}

/**
 **  Name: GenerateFileName
 **
 **  Function: 
 **
 **  Binding:  void GenerateFileName(char *FileName, char *path)
 **
 **        FileName - returned name of file to open
 **        path - path of from from MIME info
 **
 **  Returns:  
 **/

void GenerateFileName(char *FileName, char *path)
{
	char *ptr;

	strcpy(FileName, STARTIN);
	if ((path[0] == '/' || path[0] == '\\') && path[1] == 0) /* no file name */
		strcat(FileName, DEFAULT_PAGE);   /* means home page */
	else
		strcat(FileName, path);

	for (ptr = &FileName[0]; *ptr; ptr++) /* search through name */
		if ( *ptr == '/')                   /* and flip slashes */
			*ptr = '\\';

	ReqFileType = GetFileType(FileName);
}

/**
 **  Name: GetFileType
 **
 **  Function: Return file index based on file suffix
 **
 **  Binding:  int GetFileType(char *buffer)
 **
 **  Returns:  File type index
 **/

int GetFileType(char *buffer)
{
	int fcode = MaxTypes;
	char *ptr;

	if ((ptr = strrchr(buffer, '.')) != 0) /* find last period */
	{
		ptr++;

		while (fcode--)
		{
			if (!strcmp(ptr, FileSuffix[fcode]))
				return FileType[fcode];
		}
	}
	return OTHER_FILE;
}

/**
 **  Name: ParsePostRequest
 **
 **  Function: Process a post request from the client
 **
 **  Binding:  void ParsePostRequest(char *PostBuffer, int BytesRead)
 **
 **  Returns:  void
 **/

void ParsePostRequest(char *PostBuffer, int BytesRead)
{
	int CharCount, HeaderCode;
	char *ptr;
	char LineBuf[200], BodyBuf[200], response[200];
	char *RefStr;  // for later usage

	memcpy(HoldPtr, PostBuffer, BytesRead);
	HoldPtr += BytesRead;
	BytesRead += residual; /* add left over from last time */
	residual = 0;
	while (BytesRead > 0)
	{
		ParsePtr = GetHeaderLine(ParsePtr, LineBuf, &CharCount, BytesRead);

		if (CharCount == 0)    /* incomplete line - save and finish later */
		{
			residual = BytesRead;
			break;
		}
		BytesRead -= CharCount;

		if (CharCount == 2)
		{
			body = 1;

			if (BodySize <= BytesRead)
			{
				memcpy(BodyBuf, ParsePtr, BodySize);
				*(BodyBuf + BodySize) = 0;
				ProcessPost(RefStr, BodyBuf);
				sprintf(response, "%s%s%c%c", HTTP_VER, REA_NOC, CR, LF);
				NetWrite(response);
				httpstate = WaitClose0;
			}
			BytesRead = 0;
		}
		else
		{
			HeaderCode = ParseHeader(LineBuf);

			if (HeaderCode == 0)
				HeaderCode = NextHeadCode++;

			HeaderLine[HeaderCode] = ParsePtr;

			switch (HeaderCode)
			{
				case HEAD_REFER:
					break;

				case HEAD_CONTLEN:
					ptr = &LineBuf[0];
					/* find the colon */
					while (*ptr != 0x3a && *ptr != 0)
						ptr++;
					ptr++; /* skip colon */
					/* skip the white space */
					while ((*ptr == 0x20 || *ptr == 0x0b) && *ptr != 0)
						ptr++;
					if (*ptr != 0)
						BodySize  = atoi(ptr);
					else
						BodySize = 0;

					break;
			}
		}
	}
}

/**
 **  GetHeaderLine()
 **
 **  Copies a generic buffer pointer into a new buffer up to and
 **  including a carraige return/line feed. Count is passed as a
 **  pointer and max length is passed in to avoid run away pointers.
 **  The function returns the buffer pointer after the CR/LF
 **/

char *GetHeaderLine(char *BufferPtr, char *LineBuf, int *count, int MaxLength)
{
	char *holder;

	holder = BufferPtr;

	*count = 0;
	while ( *BufferPtr != CR && *BufferPtr != LF && *count < MaxLength)
	{
		*LineBuf++ = *BufferPtr++;
		*count += 1;
	}

	if (*BufferPtr == CR || *BufferPtr == LF)
	{
		BufferPtr++;
		*count += 1;
		if (*BufferPtr == CR || *BufferPtr == LF)
		{
			BufferPtr++;
			*count += 1;
		}
	}
	else      /* if count ran out, return original pointer */
	{
		BufferPtr = holder;
		*count = 0;
	}

	*LineBuf = 0; /* null termintat line */
	return BufferPtr;
}

/**
 **  Name: ParseHeader
 **
 **  Function: return code type based on header ASCII type
 **
 **  Binding:  int ParseHeader(char *buffer)
 **
 **  Returns:  index
 **/

int ParseHeader(char *buffer)
{
	int code = HEAD_MAX;
	char *ptr1, *ptr2;

	while (code >= 0)
	{
		ptr1 = buffer;
		ptr2 = headers[code];

		while (*ptr2 != 0 && tolower(*ptr1) == *ptr2)
		{
			ptr2++;
			ptr1++;
		}

		if (*ptr2 == 0)
			break;
		else
			code--;
	}
	return code;
}

/**
 **  Name: SendResponseFile
 **
 **  Function: Open requested file & return it
 **
 **  Binding:  int SendResponseFile(
 **               char *Name404,        Raw filename
 **               char *filename,       filename path
 **               int HeaderOnlyFlag)   send header only
 **
 **  Returns:  length of response
 **/

int SendResponseFile(char *Name404, char *filename, int HeaderOnlyFlag)
{
	char TmpBuffer[400];
	char *PositionIn, *PositionOut, *tmp1, *tmp2;
	int length, httpfh, LeftOver, parsed, total;

	if ((httpfh = open(filename, O_BINARY|O_RDONLY)) != -1 && NoDots(Name404))
	{
		sprintf(TmpBuffer, "%s%s%c%c", HTTP_VER, REA_OK, CR, LF);
		NetWrite(TmpBuffer);

		LeftOver = 0;
		PositionIn = &TempBuf[0];
		total = 0;
		PositionOut = &HoldBuf[0];

		while((length = read(httpfh, PositionIn, 1024)) != 0)
		{
			PositionIn = &TempBuf[0];
			length += LeftOver;
			while (1)
			{
				PositionIn = GetHeaderLine(PositionIn, TmpBuffer, &parsed, length);
				if (parsed == 0) break;
					length -= parsed;
				ProcessOutputString(TmpBuffer);
				sprintf(PositionOut, "%s%c%c", TmpBuffer, CR, LF);
				total += strlen(TmpBuffer) + 2;
				PositionOut += strlen(TmpBuffer) + 2;
			}
			LeftOver = length;
			tmp1 = &TempBuf[0];
			tmp2 = PositionIn;

			while(length--)
				*tmp1++ = *tmp2++;
			PositionIn = tmp1;
		}

		close(httpfh);

		sprintf(TmpBuffer,
			"%s %d%c%c%c%c", headers[HEAD_CONTLEN], total, CR, LF, CR, LF);
		NetWrite(TmpBuffer);

		if (!HeaderOnlyFlag)
			sock_write((void *) &HttpSocket, HoldBuf, total);
	}
	else
	{
		Send404Response(Name404);
		length = 0;
	}

	return length;
}

/**
 **  Name: SendNormalFile
 **
 **  Function: 
 **
 **  Binding:  int SendNormalFile(
 **               char *Name404,        Raw filename
 **               char *filename,       filename path
 **               int HeaderOnlyFlag)   send header only
 **
 **  Returns:  length of response
 **/

int SendNormalFile(char *Name404, char *filename, int HeaderOnlyFlag)
{
	char TmpBuffer[400];
	int length, total;

	if ((httpfh = open(filename, O_BINARY|O_RDONLY)) != -1 && NoDots(Name404))
	{
		sprintf(TmpBuffer, "%s%s%c%c", HTTP_VER, REA_OK, CR, LF);
		NetWrite(TmpBuffer);

		total = filelength(httpfh);
		sprintf(TmpBuffer,
			"%s %d%c%c%c%c", headers[HEAD_CONTLEN], total, CR, LF, CR, LF);
		NetWrite(TmpBuffer);

		if (!HeaderOnlyFlag)
			while ((length = read(httpfh, MungBuf, 1024)) != 0)
				sock_write((void *) &HttpSocket, MungBuf, length);
	}
	else
	{
		Send404Response(Name404);
		length = 0;
	}

    return length;
}

/**
 **  Name: Send404Response
 **
 **  Function: Send a 404 response if file NOT located
 **
 **  Binding:  void Send404Response(char *Name404)
 **
 **  Returns:  void
 **/

void Send404Response(char *Name404)
{
	char HdrBuffer[400], MsgBuffer[400];
	int total;

	sprintf(MsgBuffer, "%s%s%c%c", HTTP_VER, REA_NOTFOUND, CR, LF);
	NetWrite(MsgBuffer);

	// Skip the first '/' if present
	if (*Name404 == '/')
		Name404++;

	// Format the 404 OBJECT NOT FOUND page
	sprintf(MsgBuffer,
		"<body><h1>HTTP/1.0 404 Object %s Not Found%c%c</h1></body>%c%c",
		Name404, CR, LF, CR, LF);

	total = strlen(MsgBuffer);  // Build header with proper size
	sprintf(HdrBuffer,
		"%s %d%c%c%c%c", headers[HEAD_CONTLEN], total, CR, LF, CR, LF);

	NetWrite(HdrBuffer);  // Send header
	NetWrite(MsgBuffer);  // Send message

	//  Send a 404 NOT FOUND status message
	sprintf(MsgBuffer, "%s%s%c%c", HTTP_VER, REA_NOTFOUND, CR, LF);
	NetWrite(MsgBuffer);
}

/**
 **  Actual post data looks something like this:
 **
 ** POST_/_vti_bin/shtml.exe/LED.htm_HTTP/1.0
 ** Referer: http://192.9.200.214/LED.htm
 ** Connection: Keep-Alive
 ** User-Agent: Mozilla/4.5 [en] (WinNT; U)
 ** Host: 192.9.200.214
 ** Accept: image/gif, image/x-xbitmap, image/jpeg, image/pjpeg, image/png,
 ** Accept-Encoding: gzip
 ** Accept-Language: en
 ** Accept-Charset: iso-8859-1,*,utf-8
 ** Content-type:_application/x-www-form-urlencoded
 ** Content-length:_31
 **
 ** VTI-GROUP=0&D1=TOGGLE&B1=Submit
 **
 ** What gets posted to us:
 **
 **  DataButtonStr = VTI-GROUP=0&D1=TOGGLE&B1=Submit
 **     or
 **  DataButtonStr = D1=TOGGLE&B1=Submit
 **
 **  Depending on the browser and browser version
 **
 **  The reference string will look something like this:
 **  RefStr = http://192.9.200.214/LED.htm
 **
 **  This Reference String is not currently in use BUT
 **  could be used in the future for specific access
 **  validation.
 **
 **  ProcessPost(RefStr, DataButtonStr);
 **/


/**
 **  Name: ProcessPost
 **
 **  Function: Passed POST strings, parses and initiates
 **            actions if required.
 **
 **  Binding:  int ProcessPost(char *RefStr, char *DataButtonStr);
 **
 **  Returns:  1 - Processed the function
 **            0 - no recognized function keyword
 **
 **  This function will become more generic in the future
 **  For now it only parses a specific DataButtonStr type
 **  while ignoring the RefStr.
 **/

int ProcessPost(char *RefStr, char *DataButtonStr)
{
	RefStr++;  // Resolve BorlandC error/warning

	if (debugMode)
		printf("Post string: \"%s\"\n", DataButtonStr);

	while (strlen(DataButtonStr) >= 5)  // Minimum is "D1=ON"
	{
		if (DataButtonStr[0] == 'D' && DataButtonStr[1] == '1'
			&& DataButtonStr[2] == '=')
		{
			if (DataButtonStr[3] == 'T')
			{
				if (strncmp("TOGGLE", &DataButtonStr[3], 6) != 0)
					return 0;

				setLedState(LedToggle);
				return 1;
			}
			else if (DataButtonStr[3] == 'O')
			{
				if (DataButtonStr[4] == 'N')
				{
					setLedState(LedOn);
					return 1;
				}
				else if (DataButtonStr[4] == 'F')
				{
					if (DataButtonStr[5] == 'F')
					{
						setLedState(LedOff);
						return 1;
					}
					else
						return 0;
				}
				else
					return 0;
			}
		}
		else
			DataButtonStr++;
	}

	return 0;
}

/**
 **  Name:     setLedState
 **
 **  Function: Execute low level function to control LED
 **            state.  On, Off, Toggle
 **
 **  Binding:  void  setLedState(State)
 **
 **  Returns:  void
 **/

void  setLedState(int State)
{
	union REGS regs;

	regs.x.ax = 0xb010;

	if (TsSbc)
	{
		switch (State)
		{
			case LedOff:
				regs.h.bh = 0;
				break;

			case LedOn:
				regs.h.bh = 1;
				break;

			case LedToggle:
				regs.h.bh = 0x81;
		}
		int86(0x15, &regs, &regs);  /* generate int 15: ax=b010  bh=function */
	}
	else
	{
		switch (State)
		{
			case LedOff:
				printf("LED OFF\n");
				break;

			case LedOn:
				printf("LED ON\n");
				break;

			case LedToggle:
				printf("LED TOGGLE\n");
		}
	}
}

/**
 **  Name: ProcessComment
 **
 **  Function: Copy or delete a comment line.
 **
 **  Binding:
 **    void ProcessComment(char **source,
 **       char **destination,
 **       unsigned DiscardFlag)
 **
 **  Returns:  void
 **    Original addresses are updated
 **/

void ProcessComment(char **source, char **destination, unsigned discard)
{
	char *sp, *dp;

	sp = *source;  // fetch working addresses
	dp = *destination;

	while (*sp)
	{
		if (*sp == '-' && sp[1] && sp[1] == '-' && sp[2] && sp[2] == '>')
			break;

		if (discard)
			sp++;   // skip character
		else
			*dp++ = *sp++;   // copy character
	}

	if (discard)
	{
		sp++;   // skip character '-'
		sp++;   // skip character '-'
		sp++;   // skip character '>'
	}
	else
	{
		*dp++ = *sp++;   // copy character '-'
		*dp++ = *sp++;   // copy character '-'
		*dp++ = *sp++;   // copy character '>'
	}

	*source = sp;  // restore working addresses
	*destination = dp;
}

/**
 **  Name: ProcessOutputString
 **
 **  Function: Process the output string replacing comments that
 **            are Server Side Document Includes (SSDI) with the
 **            IO or DataBase References they contain
 **
 **  Binding:  char *ProcessOutputString(char *anHttpString);
 **
 **  Returns:  String with any additionaal info embedded
 **            into it in the original passed string space
 **
 **  NOTE: This function does NOT support comments that span
 **        more than 1 line
 **
 **  String input format as follows: <!--#exec tag=format -->
 **
 **  Where - tag:    the data value to fetch by name
 **          format: the format to return the tag data
 **/

char *ProcessOutputString(char *theHttpString)
{
	int sLength;
	char lclBuf[512];
	struct ssdiEntry *execCmd;
	char tagb[64];   // storage for the tag
	char fmtb[64];   // storage for the format
	char funcb[64];  // storage for output of function
	char *sPtr;     // source Pointer
	char *dPtr;     // destination Pointer (main)
	char *ldPtr;    // destination Pointer (local)
	struct ssdiEntry *pTbl;

	sLength = strlen(theHttpString);
	memcpy(lclBuf, theHttpString, sLength+1);

	sPtr = lclBuf;         // get source pointer
	dPtr = theHttpString;  // get destination pointer

	while (*sPtr)   // As long as characters left in buffer
	{
		if (*sPtr == '<' && sPtr[1] && sPtr[1] == '!' &&
			sPtr[2] && sPtr[2] == '-' && sPtr[3] && sPtr[3] == '-')
		{
			if (sPtr[4] && sPtr[4] == '#' && sPtr[5] && sPtr[5] == 'e' &&
				sPtr[6] && sPtr[6] == 'x' && sPtr[7] && sPtr[7] == 'e' &&
				sPtr[8] && sPtr[8] == 'c' && sPtr[9] && sPtr[9] == ' ')
			{
				// Validate a Server Side Document Include (SSDI) extracting
				// the tag and fmt strings and parsing them by comparison
				// with the enties in the parse table (pTbl)

				sPtr += 10;                 // skip header characters
				tagb[0] = fmtb[0] = '\0';

				// fetch TAG
				ldPtr = tagb;          // fetch tag portion
				while (*sPtr != '\0' && *sPtr != '=')
					*ldPtr++ = *sPtr++;
				*ldPtr++ = '\0';
				sPtr++;     // Skip '='

				// fetch FORMAT
				ldPtr = fmtb;  // fetch format portion
				while (*sPtr != '\0' && *sPtr != ' ' && *sPtr != '-')
					*ldPtr++ = *sPtr++;
				*ldPtr++ = '\0';

				// We've extracted the useful info, skip balance of comment
				ProcessComment(&sPtr, &dPtr, DISCARD);

				funcb[0] = '\0';  // make NULL string

				// VALIDATE TAG & FORMAT
				pTbl = ssdiTable;
				while (pTbl->tag != 0)
				{
					if ((strcmp(pTbl->tag, tagb) == 0) &&
						(strcmp(pTbl->fmt, fmtb) == 0))
					{
						(*pTbl->func)(funcb);  // If located, execute function
						break;
					}
					else
						pTbl++;  // next table entry
				}

				ldPtr = funcb;   // point @ function buffer
				if (*ldPtr)     // if function located, copy output
				{
					sLength = strlen(ldPtr);
					while (sLength--)
						*dPtr++ = *ldPtr++;  // copy function output
				}
			}
			else  // Standard comment, copy 'til end
				ProcessComment(&sPtr, &dPtr, NODISCARD);
		}
		else
			*dPtr++ = *sPtr++;   // NOT a comment, copy character
	}

	*dPtr = 0; /* terminate new string */
	return theHttpString;
}


/**
 **  Name: Scr1Dec3
 **
 **  Function: Return a 3 digit decimal value from
 **            the data in the scratchpad register of COM1
 **
 **  Binding:  void Scr1Dec3()
 **
 **  Returns:  void
 **/

void Scr1Dec3(char *aBuf)
{
	unsigned char aByte;

	aByte = inp(0x3ff);       // read COM1: scratch register
	sprintf(aBuf, "%03d", aByte);
}

/**
 **  Name: Scr1Hex2
 **
 **  Function: Return a 2 digit hexadecimal value from
 **            the data in the scratchpad register of COM1
 **
 **  Binding:  void Scr1Hex2()
 **
 **  Returns:  void
 **/

void Scr1Hex2(char *aBuf)
{
	unsigned char aByte;

	aByte = inp(0x3ff);       // read COM1: scratch register
	sprintf(aBuf, "%02x", aByte);
}

/**
 **  Name: Scr1Hex2I
 **
 **  Function: Return a 2 digit hexadecimal value from
 **            the inverted data in the scratchpad register of COM1
 **
 **  Binding:  void Scr1Hex2I()
 **
 **  Returns:  void
 **/

void Scr1Hex2I(char *aBuf)
{
	unsigned char aByte;

	aByte = ~inp(0x3ff);       // read COM1: scratch register
	sprintf(aBuf, "%02x", aByte);
}

/**
 **  Name: Scr1Bin8
 **
 **  Function: Return an 8 digit binary value from
 **            the data in the scratchpad register of COM1
 **
 **  Binding:  void Scr1Bin8()
 **
 **  Returns:  void
 **/

void Scr1Bin8(char *aBuf)
{
	unsigned char aByte;

	aByte = inp(0x3ff);       // read COM1: scratch register
	sprintf(aBuf, "%c%c%c%c%c%c%c%c",
		aByte & 0x80 ? '1' : '0', aByte & 0x40 ? '1' : '0',
		aByte & 0x20 ? '1' : '0', aByte & 0x10 ? '1' : '0',
		aByte & 0x08 ? '1' : '0', aByte & 0x04 ? '1' : '0',
		aByte & 0x02 ? '1' : '0', aByte & 0x01 ? '1' : '0');
}

/**
 **  Name: Scr1Bin8I
 **
 **  Function: Return an 8 digit binary value from
 **            the inverted data in the scratchpad register of COM1
 **
 **  Binding:  void Scr1Bin8I()
 **
 **  Returns:  void
 **/

void Scr1Bin8I(char *aBuf)
{
	unsigned char aByte;

	aByte = ~inp(0x3ff);       // read COM1: scratch register
	sprintf(aBuf, "%c%c%c%c%c%c%c%c",
		aByte & 0x80 ? '1' : '0', aByte & 0x40 ? '1' : '0',
		aByte & 0x20 ? '1' : '0', aByte & 0x10 ? '1' : '0',
		aByte & 0x08 ? '1' : '0', aByte & 0x04 ? '1' : '0',
		aByte & 0x02 ? '1' : '0', aByte & 0x01 ? '1' : '0');
}

/**
 **  Name: Time
 **
 **  Function: Return an ASCII string of the time
 **
 **  Binding:  void Time()
 **
 **  Returns:  void
 **/

void Time(char *aBuf)
{
	struct time tmPtr;

	gettime(&tmPtr);
	sprintf(aBuf, "%02d:%02d:%02d.%02d", tmPtr.ti_hour,
		tmPtr.ti_min, tmPtr.ti_sec, tmPtr.ti_hund);
}

/**
 **  Name: Date1
 **
 **  Function: Return an ASCII string of the date as mm/dd/yyyy
 **
 **  Binding:  void Date2()
 **
 **  Returns:  void
 **/

void Date1(char *aBuf)
{
	struct date dtPtr;

	getdate(&dtPtr);
	sprintf(aBuf, "%02d/%02d/%4d", dtPtr.da_mon, dtPtr.da_day, dtPtr.da_year);
}

/**
 **  Name: Date2
 **
 **  Function: Return an ASCII string of the date as:
 **            mmmmmmmm dd,yyyy  i.e. January 1, 1999
 **
 **  Binding:  void Date2()
 **
 **  Returns:  void
 **/

void Date2(char *aBuf)
{
	struct date dtPtr;

	getdate(&dtPtr);
	sprintf(aBuf, "%s %2d, %4d", monthTbl[dtPtr.da_mon],
		dtPtr.da_day, dtPtr.da_year);
}

/**
 **  Name: DateRfc1123
 **
 **  Function: Return an ASCII string of the date in an RFC 1123 compliant format:
 **		dd/Mmm/yyyy:hh:mm:ss
 **  ex:	17/Sep/1999:12:11:04
 **
 **  Binding:  void Date2()
 **
 **  Returns:  void
 **/

void DateRfc1123(char *aBuf)
{
	const char *MONTHTBL[13] = { "XXX",
			"Jan", "Feb", "Mar",
			"Apr", "May", "Jun",
			"Jul", "Aug", "Sep",
			"Oct", "Nov", "Dec" };
	
	struct date datePtr;
	struct time timePtr;

	getdate(&datePtr);
	gettime(&timePtr);
	sprintf(aBuf, "%2d/%s/%4d:%2d:%2d:%2d", datePtr.da_day, MONTHTBL[datePtr.da_mon], datePtr.da_year,
		timePtr.ti_hour, timePtr.ti_min, timePtr.ti_sec);
}

/**
 **  Name: NoDots
 **
 **  Function: Check incoming string for 'dot pairs'
 **
 **  Binding:  NoDots(char *filePath)
 **
 **  Returns:  TRUE if NO dot pairs, else FALSE
 **/

NoDots(char *filePath)
{
	while (*filePath)
	{
		if (filePath[0] == '.' && filePath[1] == '.')
			return 0;
		filePath++;
	}
	return 1;
}

/**
 **  Name: NetWrite
 **
 **  Function: Write the supplied buffer to the HTTP socket
 **
 **  Binding:  void NetWrite(char *theBuffer)
 **
 **  Returns:  void
 **/

void NetWrite(char *theBuffer)
{
	sock_write((void *) &HttpSocket, theBuffer, strlen(theBuffer));
}

