#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <direct.h>
#include <time.h>
#include <conio.h>
#include <stdarg.h>
#include <bios.h>
#include <dos.h>
#include <malloc.h>
#include <dir.h>
#include <io.h>
#include "whatami.h"
#include "tcp.h"
#include "bghttp.h"
#include "http.h"

// Local prototypes
void optparse(char *);
void printuse();
unsigned char checkSystemType();


int FileCount;
unsigned char debugMode = 0;
unsigned char notFoundPage = 0;
unsigned char TsSbc = 0;     // Indicate NOT Technologic Systems SBC


int main(int argc,char *argv[])
{
	register char *parm, ch;

	while (--argc)                  // parse command line
	{
		ch = *(parm = *++argv);
		if (ch == '-' || ch == '/')  // options preceded by '-' or '/'
			optparse(parm);
	}

	FileCount = 0;

	TsSbc = checkSystemType();  // Setup system type
	printf("Platform is %s\n", TsSbc ? "TS SBC" : "Generic System");

	sock_init();  /* Initialize Wattcp TCP/IP Kernel */
	InitHttpd();  /* Initialize Http state machine */

	while(1)
	{
		tcp_tick(NULL);  /* allow Kernel TCP functions to operate */
		httpd();         /* allow Http Daemon state machine to function */
	}
}


/**
 **  Name: checkSystemType
 **
 **  Function: Utilizes the int86:0x15 interface to determine
 **            the platform on which we are running
 **
 **  Binding:  unsigned char checkSystemType()
 **
 **  Returns:  TRUE -  Running on a Technologic Systems
 **                    Single Board Computer (SBC)
 **            FALSE - Running on a generic DOS machine
 **/

unsigned char checkSystemType()
{
	union REGS inregs, outregs;

	inregs.x.ax = 0xb000;

	int86(0x15, &inregs, &outregs);  /* generate int 15: ax=b000 */

	if (outregs.x.dx == 0x5453)
		return 1;
	else
		return 0;
}

/**
 **  Name: optparse
 **
 **  Function: parse '-' and '/' deliniated options
 **
 **  Binding:  void optparse(register char *ptr)
 **
 **            register char *ptr - pointer to an argv option
 **
 **  Returns:  void
 **/

void optparse(register char *ptr)       /* parse options */
{
	while (*++ptr)
		switch (*ptr|0x20)
		{
			case 'd':
				debugMode = 1;
				break;

			case 'n':
				notFoundPage = 1;
				break;

			default:
				printf("unknown option '%c'\n",*ptr);
			case '?':
				printuse();
				exit(0);
		}
}

static char *cmds[] = {
	VERSIONINFO,"\n"
	"Syntax:     httpd [parms]\n",
	"Function:   Start the HTTP Server daemon\n",
	"Parameters: -d or /d - Debug mode.  Print all messages\n",
	"                       received by the POST command\n",
//	"            -n or /n - Display a 404 NOT FOUND page to\n",
//	"                       the caller vs a 404 NOT FOUND response\n"
	};

/**
 **  Name: printuse
 **
 **  Function: Print the usage string
 **
 **  Binding:  void printuse()
 **
 **  Returns:  void
 **/

void printuse()
{
	register char **p = cmds;
	register char **cmdsend = cmds+(sizeof cmds)/(sizeof(char *));

	while (p < cmdsend)
		fputs(*p++,stderr);
}

